/*
 *
 *  NAME:
 *  	spica_struct_1a.h
 *
 *  PURPOSE:
 *	This module contains the declarations of the different typedef
 * 	specific to the SPICAM/SPICAV UV/IR data level 1A structure 
 *
 * Author:
 *   Nicolas Chapron
 *     CNRS | LATMOS - Laboratoire Atmospheres, Milieux, Observations Spatiales.
 *     11 boulevard d'Alembert
 *     78280 Guyancourt
 *     France
 *     Tel. 33-(0) 1 80 28 50 57
 *     Email. nicolas.chapron@latmos.ipsl.fr
 *
 *  SVN Revision:
 *      $Rev: 2366 $
 *      $Author: chapron $
 *      $Date: 2014-10-30 16:02:26 +0100 (Jeu, 30 oct 2014) $
 *
 */

#ifndef SPICA_STRUCT_1A_H
#define SPICA_STRUCT_1A_H

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "fitsio.h"

// ----- Info ------------------------------------------------------------------
typedef struct {
	
	// UV/IR
	char *instrument;
	char *obstype;
	char *begins;
	char *ends;
	int orbit;
	int seq;
	int naxis1;
	int naxis2;
	int naxis3;
	
	// UV ONLY
	int DCProc;
	int ENProc;
	int COSProc;
	int SATProc;
	int orbDCNU;
	char* File0C;
	char* data_status;
	char* geo_status;
	char* flag_status;
	char* dc_status;
	
	// IR ONLY
	char* status;
	int nb_spectrum;
        int nb_points;
        int nb_channels;
        int nb_blocks;
        int size_block;	
	
}Info;
// ----- END Info --------------------------------------------------------------

// ----- Parameters ------------------------------------------------------------
typedef struct {
	
	// UV/IR
	int peltier;
	int soiron;

	// UV ONLY
	int codeop;
	int binning;
	int ht;
	int ti;
	int x0;
	int y0;
	int slit;
	int uvsampling;
	int iron;
	int *all_ti;
	float *t_peltier;
	float *t_ccd;
	float *t_numboard;
	float *t_btboard;
	float *t_shutter;
	float *t_servboard;
	float *t_hvps;
	float *t_structure;
	
	// IR ONLY
        int uvon;
        int exit;
        int source;
        int gain;
        int dac;
        int dots;
        int ndots;
        float time;
        float gainboost;
        int dets;
        float* frequency;
        float* t_d0;
        float* t_d1;
        float* t_aotf;
        float* t_base;
        float* power_rf;
        float* pvs;
	
}Parameters;
// ----- END Parameters --------------------------------------------------------

// ----- Time_of_Records -------------------------------------------------------
typedef struct{
	
	// IR ONLY
        int* year;
        int* month;
        int* day;
        int* hour;
        int* minute;
        int* second;
        float* msecond;
	
}Time_of_Records;
// ----- END Time_of_Records ---------------------------------------------------

// ----- GeoInfo ---------------------------------------------------------------
typedef struct {
	
	// UV/IR
	char *target;
	float sunlat;
	float sunlong;
	float sundist;
	float sunls;
	float sunra;
	float sundec;
	float slitcenter[2];
	
	// UV ONLY
	char *shadowcone;
	
	// IR ONLY
        float irfov_pos[2];
        float irshift;
	
}GeoInfo;
// ----- END GeoInfo -----------------------------------------------------------

// ----- Geo_Record ------------------------------------------------------------
typedef struct {
	
	// UV/IR
	int *number;
	char **time;
	
}Geo_Record;
// ----- END Geo_Record --------------------------------------------------------

// ----- Geo_Spacecraft --------------------------------------------------------
typedef struct {
	
	// UV/IR
	float *lat;
	float *lon;
	float *alt;
	float *sza;
	float *shadowconeratio;
	
}Geo_Spacecraft;
// ----- END Geo_Spacecraft ----------------------------------------------------

// ----- Geo_Band3LOSE ---------------------------------------------------------
typedef struct {
	
	// UV ONLY
	float *sza;
	float *disttoplanetnp;
	float *pixelsize;
	float *anglelossun;
	float *phaselossun;
	float *solarincidence;
	float *solarlocaltime;
	float *anglenormalobserver;
	float *sunazimuth;
	float *distlosplanetcenter;
	float *distscplanetcenter;
	float *lat;
	float *lon;
	float *alt;
	float *ra;
	float *dec;
	
}Geo_Band3LOSE;
// ----- END Geo_Band3LOSE -----------------------------------------------------

// ----- Geo_BandCCDline -------------------------------------------------------
typedef struct {
	
	// UV ONLY
	float *lat;
	float *lon;
	float *alt;
	float *ra;
	float *dec;
	
}Geo_BandCCDline;
// ----- END Geo_BandCCDline ---------------------------------------------------

// ----- Geo_Coordinates -------------------------------------------------------
typedef struct {
	
	// UV/IR
	float *xsc_x;
	float *xsc_y;
	float *xsc_z;
	float *ysc_x;
	float *ysc_y;
	float *ysc_z;
	float *zsc_x;
	float *zsc_y;
	float *zsc_z;
	float *angle_xsc_localvert;
	float *angle_xsc_proj;
	float *p1_u;
	float *p1_v;
	float *p2_u;
	float *p2_v;
	float *p3_u;
	float *p3_v;
	float *p4_u;
	float *p4_v;
	float *p5_u;
	float *p5_v;
	
}Geo_Coordinates;
// ----- END Geo_Coordinates ---------------------------------------------------

// ----- Geo_TransMatrix -------------------------------------------------------
typedef struct {
	
	// UV/IR
	float *x_ra;
	float *x_dec;
	float *y_ra;
	float *y_dec;
	float *z_ra;
	float *z_dec;
} Geo_TransMatrix;
// ----- END Geo_TransMatrix ---------------------------------------------------

// ----- Geo -------------------------------------------------------------------
typedef struct {
	
	// UV/IR
	Geo_Record Record;
	Geo_Spacecraft Spacecraft;
	Geo_Coordinates Coordinates;
	Geo_TransMatrix TransMatrix;

	// UV ONLY
	Geo_Band3LOSE Band3;
	Geo_Band3LOSE LOSE;
	Geo_BandCCDline Band1;
	Geo_BandCCDline Band2;
	Geo_BandCCDline Band4;
	Geo_BandCCDline Band5;
	Geo_BandCCDline CCDLine;
	
	// IR ONLY
        Geo_Band3LOSE IRFOV;

}Geo;
// ----- END Geo ---------------------------------------------------------------

// ----- spica_1Afile ----------------------------------------------------------
typedef struct {
	
	// UV/IR
	float *cleandata[100];
	Info Info;
	Parameters Parameters;
	GeoInfo GeoInfo;
	Geo Geo;

	// UV ONLY	
	float *errdata[100];
	int *flag[100];
	
	// IR ONLY
	float *WL[100];
	float *DC[100];
	float *RAW[100];
	Time_of_Records Time_of_Records;

} spica_1Afile;
// ----- END spica_1Afile ------------------------------------------------------

// ----- Functions -------------------------------------------------------------

void read_CleanData(fitsfile*,spica_1Afile*,int*);
void read_ErrData(fitsfile*,spica_1Afile*,int*);
void read_Flag(fitsfile*,spica_1Afile*,int*);
void apply_Flag(spica_1Afile*);
void read_WL(fitsfile*,spica_1Afile*,int*);
void read_DC(fitsfile*,spica_1Afile*,int*);
void read_RAW(fitsfile*,spica_1Afile*,int*);
void read_Info(fitsfile*,spica_1Afile*,int*,int);
void read_Parameters(fitsfile*,spica_1Afile*,int*,int,int,int);
void read_Time_of_Records(fitsfile*,spica_1Afile*,int*);
void read_GeoInfo(fitsfile*,spica_1Afile*,int*,int);
void read_Geo_Record(fitsfile*,spica_1Afile*,int*);
void read_Geo_Spacecraft(fitsfile*,spica_1Afile*,int*,int);
void read_Geo_Band1(fitsfile*,spica_1Afile*,int*);
void read_Geo_Band2(fitsfile*,spica_1Afile*,int*);
void read_Geo_Band3(fitsfile*,spica_1Afile*,int*);
void read_Geo_Band4(fitsfile*,spica_1Afile*,int*);
void read_Geo_Band5(fitsfile*,spica_1Afile*,int*);
void read_Geo_CCDLine(fitsfile*,spica_1Afile*,int*);
void read_Geo_LOSE(fitsfile*,spica_1Afile*,int*);
void read_Geo_IRFOV(fitsfile*,spica_1Afile*,int*);
void read_Geo_Coordinates(fitsfile*,spica_1Afile*,int*);
void read_Geo_TransMatrix(fitsfile*,spica_1Afile*,int*);
void read_Geo(fitsfile*,spica_1Afile*,int*,int,int);
char *trimchars(char*);

// ----- END Functions ---------------------------------------------------------

#endif SPICA_STRUCT_1A_H		/* SPICA_STRUCT_1A_H */
