 !
 !  NAME:
 !  	spica_read_1a_mod.f
 !
 !  PURPOSE:
 !	This module contains the function spica_read_1a to read the SPICAM and SPICAV UV/IR data in level 1A.
 !      The function returns a FORTRAN derived data type (struc_1AFile) corresponding to the FITS file. 
 !	To get more informations about the content of the returned data type, please read the PDF file 
 !	named "Using-SPICAM-V-1A-data.pdf" or the CONTENT_UV.TXT/CONTENT_IR.TXT
 !
 ! Inputs:
 ! 	spica_read_1a:
 !      - filename           [string] : full path and name of the 1A file
 !      - (OPTIONAL) mode    [Int]    : output messages mode, 0:output messages disabled; 1: output messages enabled
 !					Default value=0
 !      - (OPTIONAL) no_mask [Int]    : 0:The flag mask will be appliked, 1:The flag mask will NOT be applied
 !				        This parameter should be only defined for UV data
 !					Default value=0
 !
 ! Output:
 !      - struct_1A [struct_1Afile]: 1A file content structure
 !
 ! Library:
 !      cfitsio: http://heasarc.gsfc.nasa.gov/fitsio/
 !      To compile this module: - Import the static library "libcfitsio.a" in your system library path
 !                              - Add the option "-I cfitsio" or "out libcfitsio.a" in you compiler command line
 !
 ! Author:
 !   Nicolas Chapron
 !     CNRS | LATMOS - Laboratoire Atmospheres, Milieux, Observations Spatiales.
 !     11 boulevard d'Alembert
 !     78280 Guyancourt
 !     France
 !     Tel. 33-(0) 1 80 28 50 57
 !     Email. nicolas.chapron@latmos.ipsl.fr
 !
 !  SVN Revision:
 !      $Rev: 2365 $
 !      $Author: chapron $
 !      $Date: 2014-10-30 16:01:18 +0100 (Jeu, 30 oct 2014) $
 !
 
module spica_read_1a_mod
	
	use spica_struct_1a_mod_func
	implicit none
	
	contains
	
	    ! ----- spica_read_1A -------------------------------------------------------------------------------------------------------------------
	    type(struc_1AFile) function spica_read_1A(filename,mode_,no_mask_)
		
		character(len=30) errmsg
		character(len=2048) filename
		integer :: unit, status, blocksize, pos_uv, mode, no_mask
		integer, intent(in), optional :: mode_
		integer, intent(in), optional :: no_mask_
		type(struc_1AFile) :: struct_1A
		    
		! Init  
		if(present(mode_)) then
			mode = mode_
		else
			mode = 0
		endif
		if(present(no_mask_)) then
			no_mask = no_mask_
		else
			no_mask = 0
		endif
		
		status = 0
		
		! Open the FITS file
		call ftgiou(unit,status)
		call ftgerr(status,errmsg)
		call ftopen(unit,filename,0,blocksize,status)
		
		if(status .ne. 0) then
			call ftgerr(status,errmsg)
			print *,errmsg
		else
			! Read file
			pos_uv = index(filename,"1AU",.true.)

			if(pos_uv .ne. 0) then
				if(mode .ne. 0) then
					print *,"1A UV reading process started..."
				endif
				struct_1A = spica_read_1AUV(filename,unit,mode,no_mask)
			else
				if(mode .ne. 0) then
					print *,"1A IR reading process started..."
				endif
				struct_1A = spica_read_1AIR(filename,unit,mode,no_mask)
			endif
			
			! Close FITS file
			call ftclos(unit, status)
			call ftfiou(unit, status)	
			
			spica_read_1A = struct_1A
	
		endif

	    end function spica_read_1A
	    ! ----- END spica_read_1A ---------------------------------------------------------------------------------------------------------------

	    
	    ! ----- spica_read_1AUV -----------------------------------------------------------------------------------------------------------------
	    type(struc_1AFile) function spica_read_1AUV(filename,unit,mode,no_mask)
		integer :: unit,readwrite,err,status, blocksize, mode, no_mask
		integer :: check_clean, check_err, check_flag, check_info, &
			    check_param, check_geoinfo, check_geo, check_all
		integer :: uvir
		character(len=2048) filename
		character(len=30) errmsg
		type(struc_Info) :: Info
		type(struc_Parameters) :: Parameters
		type(struc_GeoInfo) :: GeoInfo
		type(struc_Geo) :: Geo
		type(struc_1AFile) :: struct_1AUV
		real, dimension(:,:,:), allocatable :: CleanData
		real, dimension(:,:,:), allocatable :: ErrData
		integer, dimension(:,:,:), allocatable :: Flag
	
		check_info = 1
		check_param = 1
		check_geoinfo = 1
		check_clean = 1
		check_err = 1
		check_flag = 1
		check_geo = 1
	
		uvir = 0
		
		    Info = readInfo(unit,status,uvir)
		    if(status .eq. 0) then
			if(mode .ne. 0) then
			    print *,'Info successfully read'
			endif
			struct_1AUV%Info = Info
			check_info = 0
		    endif
	
		    Parameters = readParameters(unit,status,0,0,uvir)
		    if(status .eq. 0) then
			if(mode .ne. 0) then
			    print *,'Parameters successfully read'
			endif
			struct_1AUV%Parameters = Parameters
			check_param = 0
		    endif
	
		    !!! Geo Info
		    GeoInfo = readGeoInfo(unit,status,uvir)
		    if(status .eq. 0) then
			if(mode .ne. 0) then
			    print *,'Geo Info successfully read'
			endif
			struct_1AUV%GeoInfo = GeoInfo
			check_geoinfo = 0
		    endif
		    
		    !!! Geo Data
		    Geo = readGeo(unit,status,Parameters%codeop,mode,uvir)
		    if(status .eq. 0) then
			if(mode .ne. 0) then
			    print *,'Geo Data successfully read'
			endif
			struct_1AUV%Geo = Geo
			check_geo = 0
		    endif
		    
		    !!! Clean Data
		    allocate(CleanData(Info%naxis1,Info%naxis2,Info%naxis3),stat=err)
		    if(err .ne. 0) then
			print  *,'Array allocation error'
			call exit()
		    endif
	
		    allocate(struct_1AUV%CleanData(Info%naxis1,Info%naxis2,Info%naxis3),stat=err)
		    if(err .ne. 0) then
			print  *,'Array allocation error'
			call exit()
		    endif
	
		    call readCleanData(unit,status,Info%naxis1,Info%naxis2,Info%naxis3,CleanData)
		    if(status .eq. 0) then
			if(mode .ne. 0) then
			    print *,'Clean Data successfully read'
			endif
			struct_1AUV%CleanData = CleanData
			check_clean = 0
		    endif
		    
		    !!! Error Data
		    allocate(ErrData(Info%naxis1,Info%naxis2,Info%naxis3),stat=err)
		    if(err .ne. 0) then
			print  *,'Array allocation error'
			call exit()
		    endif
	
		    allocate(struct_1AUV%ErrData(Info%naxis1,Info%naxis2,Info%naxis3),stat=err)
		    if(err .ne. 0) then
			print  *,'Array allocation error'
			call exit()
		    endif
	
		    call readErrData(unit,status,Info%naxis1,Info%naxis2,Info%naxis3,ErrData)
		    if(status .eq. 0) then
			if(mode .ne. 0) then
			    print *,'Err Data successfully read'
			endif
			struct_1AUV%ErrData = ErrData
			check_err = 0
		    endif
	
		    !!! Flag Data
		    allocate(FLag(Info%naxis1,Info%naxis2,Info%naxis3),stat=err)
		    if(err .ne. 0) then
			print  *,'Array allocation error'
			call exit()
		    endif
	
		    allocate(struct_1AUV%FLag(Info%naxis1,Info%naxis2,Info%naxis3),stat=err)
		    if(err .ne. 0) then
			print  *,'Array allocation error'
			call exit()
		    endif
	
		    call readFlag(unit,status,Info%naxis1,Info%naxis2,Info%naxis3,Flag)
		    if(status .eq. 0) then
			if(mode .ne. 0) then
			    print *,'Flag Data successfully read'
			endif
			struct_1AUV%FLag = FLag
			check_flag = 0
		    endif
		    
		    if(no_mask .eq. 0) then
			call applyFlag(status,Info%naxis1,Info%naxis2,Info%naxis3,Flag,CleanData)
			if(mode .ne. 0) then
			    print *,'Flag mask applied'
			endif
			struct_1AUV%CleanData = CleanData
		    endif
		    
		    check_all = check_info + check_param + check_geoinfo + check_clean + check_geo + check_err
	
		    if(check_all .eq. 0 .and. mode .ne. 0) then
			print *,'The file '//file_basename(filename)//'was read succesfully!'
		    endif
	
		    spica_read_1AUV = struct_1AUV
	
	    end function spica_read_1AUV
	    ! ----- END spica_read_1AUV -------------------------------------------------------------------------------------------------------------
	    
	    
	    ! ----- spica_read_1AIR -----------------------------------------------------------------------------------------------------------------
	    type(struc_1AFile) function spica_read_1AIR(filename,unit,mode,no_mask)
		integer :: unit,readwrite,err,status, blocksize, mode, no_mask
		integer :: check_info, check_param, check_geoinfo, check_geo, &
		           check_clean, check_wl, check_dc,check_raw, check_tor, check_all
		integer :: uvir
		character(len=2048) filename
		character(len=30) errmsg
		type(struc_1AFile) :: struct_1AIR
		type(struc_Info) :: Info
		type(struc_Parameters) :: Parameters
		type(struc_Time_of_Records) :: Time_of_Records
		type(struc_GeoInfo) :: GeoInfo
		type(struc_Geo) :: Geo
		real, dimension(:,:,:), allocatable :: CleanData
		real, dimension(:,:,:), allocatable :: WAVELENGTH
		real, dimension(:,:,:), allocatable :: DC
		real, dimension(:,:,:), allocatable :: RAW
	
		check_info = 1
		check_param = 1
		check_geoinfo = 1
		check_clean = 1
		check_wl = 1
		check_dc = 1
		check_raw = 1
		check_geo = 1
		check_tor = 1
	
		uvir = 1
		
		    !!! Info
		    Info = readInfo(unit,status,uvir)
		    if(status .eq. 0) then
			if(mode .ne. 0) then
			    print *,'HDU1 HEADER successfully read'
			endif
			struct_1AIR%Info = Info
			check_info = 0
		    else
		    	call exit()
		    endif
		    
		    !!! CleanData
		    allocate(CleanData(Info%naxis1,Info%naxis2,Info%naxis3),stat=err)
		    if(err .ne. 0) then
			print  *,'Array allocation error'
			call exit()
		    endif
	
		    allocate(struct_1AIR%CleanData(Info%naxis1,Info%naxis2,Info%naxis3),stat=err)
		    if(err .ne. 0) then
			print  *,'Array allocation error'
			call exit()
		    endif
	
		    call readCleanData(unit,status,Info%naxis1,Info%naxis2,Info%naxis3,CleanData)
		    if(status .eq. 0) then
			if(mode .ne. 0) then
			    print *,'HDU1 (CleanData) successfully read'
			endif
			struct_1AIR%CleanData = CleanData
			check_clean = 0
		    else
		    	call exit()
		    endif
		    
		    !!! WAVELENGTH
		    allocate(WAVELENGTH(Info%naxis1,Info%naxis2,Info%naxis3),stat=err)
		    if(err .ne. 0) then
			print  *,'Array allocation error'
			call exit()
		    endif
	
		    allocate(struct_1AIR%WAVELENGTH(Info%naxis1,Info%naxis2,Info%naxis3),stat=err)
		    if(err .ne. 0) then
			print  *,'Array allocation error'
			call exit()
		    endif
	
		    call readWL(unit,status,Info%naxis1,Info%naxis2,Info%naxis3,WAVELENGTH)
		    if(status .eq. 0) then
			if(mode .ne. 0) then
			    print *,'HDU2 (WAVELENGTH) successfully read'
			endif
			struct_1AIR%WAVELENGTH = WAVELENGTH
			check_wl = 0
		    else
		    	call exit()
		    endif
		    
		    !!! DC
		    allocate(DC(Info%naxis1,Info%naxis2,Info%naxis3),stat=err)
		    if(err .ne. 0) then
			print  *,'Array allocation error'
			call exit()
		    endif
	
		    allocate(struct_1AIR%DC(Info%naxis1,Info%naxis2,Info%naxis3),stat=err)
		    if(err .ne. 0) then
			print  *,'Array allocation error'
			call exit()
		    endif
	
		    call readDC(unit,status,Info%naxis1,Info%naxis2,Info%naxis3,DC)
		    if(status .eq. 0) then
			if(mode .ne. 0) then
			    print *,'HDU3 (DC) successfully read'
			endif
			struct_1AIR%DC = DC
			check_dc = 0
		    else
		    	call exit()
		    endif
	
		    !!! RAW
		    allocate(RAW(Info%naxis1,Info%naxis2,Info%naxis3),stat=err)
		    if(err .ne. 0) then
			print  *,'Array allocation error'
			call exit()
		    endif
	
		    allocate(struct_1AIR%RAW(Info%naxis1,Info%naxis2,Info%naxis3),stat=err)
		    if(err .ne. 0) then
			print  *,'Array allocation error'
			call exit()
		    endif
	
		    call readRAW(unit,status,Info%naxis1,Info%naxis2,Info%naxis3,RAW)
		    if(status .eq. 0) then
			if(mode .ne. 0) then
			    print *,'HDU4 (RAW) successfully read'
			endif
			struct_1AIR%RAW = RAW
			check_raw = 0
		    else
		    	call exit()
		    endif
		
		    !!! Time of Records
		    Time_of_Records = readTime_of_Records(unit,status)
		    if(status .eq. 0) then
			if(mode .ne. 0) then
			    print *,'HDU5 (Time_of_Records) successfully read'
			endif
			struct_1AIR%Time_of_Records = Time_of_Records
			check_geo = 0
		    else
		    	call exit()
		    endif
		    
		    !!! Parameters
		    Parameters = readParameters(unit,status,Info%nb_spectrum,Info%nb_points,uvir)
		    if(status .eq. 0) then
			if(mode .ne. 0) then
			    print *,'HDU6 (Functional_Parameters) successfully read'
			endif
			struct_1AIR%Parameters = Parameters
			check_param = 0
		    else
		    	call exit()
		    endif
		    
		    !!! Geo Info
		    GeoInfo = readGeoInfo(unit,status,uvir)
		    if(status .eq. 0) then
			if(mode .ne. 0) then
			    print *,'HDU7 (Geo_Info) successfully read'
			endif
			struct_1AIR%GeoInfo = GeoInfo
			check_geoinfo = 0
		    else
		    	call exit()
		    endif
		    
		    !!! Geo Data
		    Geo = readGeo(unit,status,0,mode,uvir)
		    if(status .eq. 0) then
			if(mode .ne. 0) then
			    print *,'HDU8-12 (Geo Data) successfully read'
			endif
			struct_1AIR%Geo = Geo
			check_geo = 0
		    else
		    	call exit()
		    endif
		    		    

		    check_all = check_info + check_param + check_geoinfo + check_geo + check_clean + check_dc + check_raw + check_wl
	
		    if(check_all .eq. 0 .and. mode .ne. 0) then
			print *,'The file '//file_basename(filename)//'was read succesfully!'
		    endif
	
		    spica_read_1AIR = struct_1AIR
	
	    end function spica_read_1AIR
	    ! ----- END spica_read_1AIR -------------------------------------------------------------------------------------------------------------

end module spica_read_1a_mod

