 !
 !  NAME:
 !  	spica_struct_1a_mod_func.f
 !
 !  PURPOSE:
 !	This module contains the functions used to read all the FITS extension 
 !	data and header of the SPICAM/SPICAV UV/IR data level 1A
 !
 ! Library:
 !      cfitsio: http://heasarc.gsfc.nasa.gov/fitsio/
 !      To compile this module: - Import the static library "libcfitsio.a" in your system library path
 !                              - Add the option "-I cfitsio" or "out libcfitsio.a" in you compiler command line
 !
 ! Author:
 !   Nicolas Chapron
 !     CNRS | LATMOS - Laboratoire Atmospheres, Milieux, Observations Spatiales.
 !     11 boulevard d'Alembert
 !     78280 Guyancourt
 !     France
 !     Tel. 33-(0) 1 80 28 50 57
 !     Email. nicolas.chapron@latmos.ipsl.fr
 !
 !  SVN Revision:
 !      $Rev: 2365 $
 !      $Author: chapron $
 !      $Date: 2014-10-30 16:01:18 +0100 (Jeu, 30 oct 2014) $
 !
 
module spica_struct_1a_mod_func

    ! Import type declaration
    use spica_struct_1a_mod_type
    
    implicit none

! ----- Functions ----------------------------------------------------------------------------------------------------------------------------------------------
! --------------------------------------------------------------------------------------------------------------------------------------------------------------

contains


    ! ----- readInfo -----------------------------------------------------------------------------------------------------------------------
    ! Read primary header
    type(struc_Info) function readInfo(unit_,status,uvir)
        
    	integer :: unit_,status,status_all,uvir
        character(len=30) :: errmsg
    	character(len=80) :: comm
        type(struc_Info) :: Info

        status_all = 0

        ! UV/IR
        call ftgkys(unit_,'INSTRU',Info%Instrument,comm,status)
        status_all = status_all + status
        call ftgkys(unit_,'OBSTYPE',Info%obstype,comm,status)
        status_all = status_all + status
        call ftgkys(unit_,'BEGINS',Info%begins,comm,status)
        status_all = status_all + status
        call ftgkys(unit_,'ENDS',Info%ends,comm,status)
        status_all = status_all + status
        call ftgkyj(unit_,'ORBIT',Info%orbit,comm,status)
        status_all = status_all + status
        call ftgkyj(unit_,'SEQ_NB',Info%seq,comm,status)
        status_all = status_all + status
        call ftgkyj(unit_,'NAXIS1',Info%naxis1,comm,status)
        status_all = status_all + status
        call ftgkyj(unit_,'NAXIS2',Info%naxis2,comm,status)
        status_all = status_all + status
        call ftgkyj(unit_,'NAXIS3',Info%naxis3,comm,status)
        status_all = status_all + status
        
        ! UV ONLY
        if(uvir .eq. 0) then
        	call ftgkyj(unit_,'DCProc',Info%DCProc,comm,status)
        	status_all = status_all + status
        	call ftgkyj(unit_,'ENProc',Info%ENProc,comm,status)
        	status_all = status_all + status
        	call ftgkyj(unit_,'COSProc',Info%COSProc,comm,status)
        	status_all = status_all + status
		call ftgkyj(unit_,'SATProc',Info%SATProc,comm,status)
		status_all = status_all + status
		call ftgkyj(unit_,'ORBDCNU',Info%orbDCNU,comm,status)
		status_all = status_all + status
		call ftgkys(unit_,'DATA_0C',Info%File0C,comm,status)
		status_all = status_all + status		
		call ftgkys(unit_,'DATA_SS',Info%Data_status,comm,status)
		status_all = status_all + status
		call ftgkys(unit_,'GEO_SS',Info%Geo_status,comm,status)
		status_all = status_all + status
		call ftgkys(unit_,'FLAG_SS',Info%Flag_status,comm,status)
		status_all = status_all + status
		call ftgkys(unit_,'DC_SS',Info%DC_status,comm,status)
		status_all = status_all + status
	endif
	
	! IR ONLY
        if(uvir .eq. 1) then
        	call ftgkys(unit_,'STATUS',Info%status,comm,status)
        	status_all = status_all + status
		call ftgkyj(unit_,'NB_SPECT',Info%nb_spectrum,comm,status)
		status_all = status_all + status        	
		call ftgkyj(unit_,'NB_POINT',Info%nb_points,comm,status)
		status_all = status_all + status        	
		call ftgkyj(unit_,'NB_CHANN',Info%nb_channels,comm,status)
		status_all = status_all + status        	
		call ftgkyj(unit_,'NB_BLOCK',Info%nb_blocks,comm,status)
		status_all = status_all + status        	
		call ftgkyj(unit_,'SIZE_BL',Info%size_block,comm,status)
		status_all = status_all + status
        endif
		
        status = status_all
        
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readInfo: '//errmsg
        endif
        
        readInfo = Info

    end function readInfo
    ! ----- END readInfo -----------------------------------------------------------------------------------------------------------------------

    
    ! ----- readParameters ---------------------------------------------------------------------------------------------------------------------
    ! Read Functional_Parameters
    type(struc_Parameters) function readParameters(unit_,status,nb_spectrum,nb_points,uvir)
        integer :: unit_,err,status,status_all,nullval,frow,felem,naxis1,naxis2,err_alloc,uvir
        integer :: nb_spectrum,nb_points
        character(len=80) :: comm
        character(len=30) :: errmsg
        logical anyf
        type(struc_Parameters) :: Parameters

        ! ------ Init ------
        status_all = 0
        err_alloc = 0
        nullval = 0
        frow = 1
        felem = 1

        ! ------ Move to the HDU ------
        call ftmnhd(unit_,2,'Functional_Parameters',0,status) !ftmhd type 2 = bin table
        status_all = status_all + status
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readParameters: '//errmsg
        endif

        ! ------ Array Allocation ------
        call ftgkyj(unit_,'NAXIS1',naxis1,comm,status)
        call ftgkyj(unit_,'NAXIS2',naxis2,comm,status)
        
        ! UV ONLY
        if(uvir .eq. 0) then
		allocate(Parameters%all_ti(naxis2),stat=err)
		err_alloc = err_alloc + err
		allocate(Parameters%t_peltier(naxis2),stat=err)
		err_alloc = err_alloc + err
		allocate(Parameters%t_ccd(naxis2),stat=err)
		err_alloc = err_alloc + err
		allocate(Parameters%t_numboard(naxis2),stat=err)
		err_alloc = err_alloc + err
		allocate(Parameters%t_btboard(naxis2),stat=err)
		err_alloc = err_alloc + err
		allocate(Parameters%t_shutter(naxis2),stat=err)
		err_alloc = err_alloc + err
		allocate(Parameters%t_servboard(naxis2),stat=err)
		err_alloc = err_alloc + err
		allocate(Parameters%t_hvps(naxis2),stat=err)
		err_alloc = err_alloc + err
		allocate(Parameters%t_structure(naxis2),stat=err)
		err_alloc = err_alloc + err
        endif
        
        ! IR ONLY
        if(uvir .eq. 1) then
		allocate(Parameters%frequency(nb_points),stat=err)
		err_alloc = err_alloc + err
		allocate(Parameters%t_d0(nb_spectrum),stat=err)
		err_alloc = err_alloc + err
		allocate(Parameters%t_d1(nb_spectrum),stat=err)
		err_alloc = err_alloc + err
		allocate(Parameters%t_aotf(nb_spectrum),stat=err)
		err_alloc = err_alloc + err
		allocate(Parameters%t_base(nb_spectrum),stat=err)
		err_alloc = err_alloc + err
		allocate(Parameters%power_rf(nb_spectrum),stat=err)
		err_alloc = err_alloc + err
		allocate(Parameters%pvs(nb_spectrum),stat=err)
		err_alloc = err_alloc + err
        endif
        
        if(err_alloc .ne. 0) then
            print  *,'readParameters: Array allocation error'
            call exit()
        endif

        ! ------ Read parameters ------
        
        ! UV/IR
        call ftgkyj(unit_,'PELTIER',Parameters%peltier,comm,status)
        status_all = status_all + status
	call ftgkyj(unit_,'SOIR_ON',Parameters%soiron,comm,status)
	status_all = status_all + status
		
        ! UV ONLY
        if(uvir .eq. 0) then
		call ftgkyj(unit_,'CODEOP',Parameters%codeop,comm,status)
		status_all = status_all + status
		call ftgkyj(unit_,'BINNING',Parameters%binning,comm,status)
		status_all = status_all + status
		call ftgkyj(unit_,'TI',Parameters%ti,comm,status)
		status_all = status_all + status
		call ftgkyj(unit_,'HT',Parameters%ht,comm,status)
		status_all = status_all + status
		call ftgkyj(unit_,'X0',Parameters%x0,comm,status)
		status_all = status_all + status
		call ftgkyj(unit_,'Y0',Parameters%y0,comm,status)
		status_all = status_all + status
		call ftgkyj(unit_,'SLIT',Parameters%slit,comm,status)
		status_all = status_all + status
		call ftgkyj(unit_,'UVSAMPL',Parameters%uvsampling,comm,status)
		status_all = status_all + status
		call ftgkyj(unit_,'IR_ON',Parameters%iron,comm,status)
		status_all = status_all + status
        endif
        
        ! IR ONLY
        if(uvir .eq. 1) then
		call ftgkyj(unit_,'UVON',Parameters%uvon,comm,status)
		status_all = status_all + status        
		call ftgkyj(unit_,'EXIT',Parameters%exit,comm,status)
		status_all = status_all + status        
		call ftgkyj(unit_,'SOURCE',Parameters%source,comm,status)
		status_all = status_all + status     
		call ftgkyj(unit_,'GAIN',Parameters%gain,comm,status)
		status_all = status_all + status       
		call ftgkyj(unit_,'DAC',Parameters%dac,comm,status)
		status_all = status_all + status                
		call ftgkyj(unit_,'DOTS',Parameters%dots,comm,status)
		status_all = status_all + status                
		call ftgkyj(unit_,'NDOTS',Parameters%ndots,comm,status)
		status_all = status_all + status                
		call ftgkye(unit_,'TIME',Parameters%time,comm,status)
		status_all = status_all + status
		call ftgkye(unit_,'GAINBST',Parameters%gainboost,comm,status)
		status_all = status_all + status
		call ftgkyj(unit_,'DETS',Parameters%dets,comm,status)
		status_all = status_all + status                
	endif
        
        ! ------ Read table data ------
        
        ! UV ONLY
        if(uvir .eq. 0) then
		call ftgcvj(unit_,1,frow,felem,naxis2,nullval,Parameters%all_ti,anyf,status)
		status_all = status_all + status
		call ftgcve(unit_,2,frow,felem,naxis2,nullval,Parameters%t_peltier,anyf,status)
		status_all = status_all + status
		call ftgcve(unit_,3,frow,felem,naxis2,nullval,Parameters%t_ccd,anyf,status)
		status_all = status_all + status
		call ftgcve(unit_,4,frow,felem,naxis2,nullval,Parameters%t_numboard,anyf,status)
		status_all = status_all + status
		call ftgcve(unit_,5,frow,felem,naxis2,nullval,Parameters%t_btboard,anyf,status)
		status_all = status_all + status
		call ftgcve(unit_,6,frow,felem,naxis2,nullval,Parameters%t_shutter,anyf,status)
		status_all = status_all + status
		call ftgcve(unit_,7,frow,felem,naxis2,nullval,Parameters%t_servboard,anyf,status)
		status_all = status_all + status
		call ftgcve(unit_,8,frow,felem,naxis2,nullval,Parameters%t_hvps,anyf,status)
		status_all = status_all + status
		call ftgcve(unit_,9,frow,felem,naxis2,nullval,Parameters%t_structure,anyf,status)
		status_all = status_all + status
        endif
        
        ! IR ONLY
        if(uvir .eq. 1) then
		call ftgcve(unit_,1,frow,felem,nb_points,nullval,Parameters%frequency,anyf,status)
		status_all = status_all + status        
		call ftgcve(unit_,2,frow,felem,nb_spectrum,nullval,Parameters%t_d0,anyf,status)
		status_all = status_all + status        
		call ftgcve(unit_,3,frow,felem,nb_spectrum,nullval,Parameters%t_d1,anyf,status)
		status_all = status_all + status        
		call ftgcve(unit_,4,frow,felem,nb_spectrum,nullval,Parameters%t_aotf,anyf,status)
		status_all = status_all + status        
		call ftgcve(unit_,5,frow,felem,nb_spectrum,nullval,Parameters%t_base,anyf,status)
		status_all = status_all + status        
		call ftgcve(unit_,6,frow,felem,nb_spectrum,nullval,Parameters%power_rf,anyf,status)
		status_all = status_all + status        
		call ftgcve(unit_,7,frow,felem,nb_spectrum,nullval,Parameters%pvs,anyf,status)
		status_all = status_all + status                
        endif
        
        status = status_all
        
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readParameters: '//errmsg
        endif

        readParameters = Parameters

    end function readParameters
    ! ----- END readParameters -----------------------------------------------------------------------------------------------------------------

    
    ! ----- readTime_of_Records ----------------------------------------------------------------------------------------------------------------
    ! Read Time of Records
    type(struc_Time_of_Records) function readTime_of_Records(unit_,status)
        integer :: unit_,nullval,frow,felem,err,err_alloc,status,pos,status_all,naxis2,uvir
        character(len=80) :: comm
        character(len=30) :: errmsg
        logical anyf
        type(struc_Time_of_Records) :: Time_of_Records        
        
        ! ------ Init ------
        status_all = 0
        err_alloc = 0
        nullval = 0
        frow = 1
        felem = 1        
        
        ! ------ Move to the HDU ------
        call ftmnhd(unit_,2,'Time_of_Records',0,status) !ftmhd type 2 = bin table
        status_all = status_all + status
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readTime_of_Records: '//errmsg
        endif
        
        ! ------ Array Allocation ------
        
        call ftgkyj(unit_,'NAXIS2',naxis2,comm,status)
        
        allocate(Time_of_Records%year(naxis2),stat=err)
	err_alloc = err_alloc + err        
        allocate(Time_of_Records%month(naxis2),stat=err)
	err_alloc = err_alloc + err        
        allocate(Time_of_Records%day(naxis2),stat=err)
	err_alloc = err_alloc + err        
        allocate(Time_of_Records%hour(naxis2),stat=err)
	err_alloc = err_alloc + err        
        allocate(Time_of_Records%minute(naxis2),stat=err)
	err_alloc = err_alloc + err        
        allocate(Time_of_Records%second(naxis2),stat=err)
	err_alloc = err_alloc + err        
        allocate(Time_of_Records%msecond(naxis2),stat=err)
	err_alloc = err_alloc + err    
	
        ! ------ Read table data ------
	
	call ftgcvj(unit_,1,frow,felem,naxis2,nullval,Time_of_Records%year,anyf,status)
	status_all = status_all + status
	call ftgcvj(unit_,2,frow,felem,naxis2,nullval,Time_of_Records%month,anyf,status)
	status_all = status_all + status
	call ftgcvj(unit_,3,frow,felem,naxis2,nullval,Time_of_Records%day,anyf,status)
	status_all = status_all + status
	call ftgcvj(unit_,4,frow,felem,naxis2,nullval,Time_of_Records%hour,anyf,status)
	status_all = status_all + status
	call ftgcvj(unit_,5,frow,felem,naxis2,nullval,Time_of_Records%minute,anyf,status)
	status_all = status_all + status
	call ftgcvj(unit_,6,frow,felem,naxis2,nullval,Time_of_Records%second,anyf,status)
	status_all = status_all + status
	call ftgcve(unit_,7,frow,felem,naxis2,nullval,Time_of_Records%msecond,anyf,status)
	status_all = status_all + status
	
        status = status_all
        
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readTime_of_Records: '//errmsg
        endif
        
        readTime_of_Records = Time_of_Records	
    
    end function readTime_of_Records
    ! ----- END readTime_of_Records ------------------------------------------------------------------------------------------------------------
    
    
    ! ----- readGeoInfo -----------------------------------------------------------------------------------------------------------------------
    ! Read GEO Infos
    type(struc_GeoInfo) function readGeoInfo(unit_,status,uvir)
        integer :: unit_, nullval, status, anyf, frow, felem, pos, status_all, uvir
        character(len=80) :: comm
        character(len=16) :: slitc,irfovpos
        character(len=30) :: errmsg
        type(struc_GeoInfo) :: GeoInfo

        ! ------ Init ------
        status_all = 0

	! Move to the HDU
	call ftmnhd(unit_,2,'Geo_Record',0,status) !ftmhd type 2 = bin table
	status_all = status_all + status
	if(status .ne. 0) then
	    call ftgerr(status,errmsg)
	    print *,'readGeoInfo: '//errmsg
	endif  
	
        ! ------ Read parameters ------
	! UV/IR
        call ftgkys(unit_,'SLIT_C',slitc,comm,status)
	status_all = status_all + status
	pos = index(slitc," ",.true.)
	read(slitc(1:pos),*) GeoInfo%slitcenter(1)
	read(slitc(pos:len(slitc)),*) GeoInfo%slitcenter(2)        	
	
        ! UV ONLY
        if(uvir .eq. 0) then
		call ftgkys(unit_,'TARGET',GeoInfo%target,comm,status)
		status_all = status_all + status
		call ftgkys(unit_,'CONE',GeoInfo%shadowcone,comm,status)
		status_all = status_all + status

		call ftgkye(unit_,'SUNLAT',GeoInfo%sunlat,comm,status)
		status_all = status_all + status
		call ftgkye(unit_,'SUNLONG',GeoInfo%sunlong,comm,status)
		status_all = status_all + status
		call ftgkye(unit_,'SUNDIST',GeoInfo%sundist,comm,status)
		status_all = status_all + status
		call ftgkye(unit_,'SUNLS',GeoInfo%sunls,comm,status)
		status_all = status_all + status
		call ftgkye(unit_,'SUNRA',GeoInfo%sunra,comm,status)
		status_all = status_all + status
		call ftgkye(unit_,'SUNDEC',GeoInfo%sundec,comm,status)
		status_all = status_all + status
        endif
        
        ! IR ONLY
        if(uvir .eq. 1) then
		call ftgkys(unit_,'TARGET',GeoInfo%target,comm,status)
		status_all = status_all + status
		call ftgkye(unit_,'SUNLAT',GeoInfo%sunlat,comm,status)
		status_all = status_all + status
		call ftgkye(unit_,'SUNLONG',GeoInfo%sunlong,comm,status)
		status_all = status_all + status
		call ftgkye(unit_,'SUNDIST',GeoInfo%sundist,comm,status)
		status_all = status_all + status
		call ftgkye(unit_,'SUNLS',GeoInfo%sunls,comm,status)
		status_all = status_all + status
		call ftgkye(unit_,'SUNRA',GeoInfo%sunra,comm,status)
		status_all = status_all + status
		call ftgkye(unit_,'SUNDEC',GeoInfo%sundec,comm,status)
		status_all = status_all + status        
		call ftgkys(unit_,'IRSHIFT',GeoInfo%irshift,comm,status)
		status_all = status_all + status
		call ftgkys(unit_,'IRFOVPOS',irfovpos,comm,status)
		status_all = status_all + status
		
		pos = index(irfovpos," ",.true.)
		read(irfovpos(1:pos),*) GeoInfo%irfov_pos(1)
		read(irfovpos(pos:len(irfovpos)),*) GeoInfo%irfov_pos(2)     
        endif        
        
        status = status_all
        
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readGeoInfo: '//errmsg
        endif
        
        readGeoInfo = GeoInfo

    end function readGeoInfo
    ! ----- END readGeoInfo -------------------------------------------------------------------------------------------------------------------
    
    
    ! ----- readGeo_Record ---------------------------------------------------------------------------------------------------------------------
    ! Read GEO Record
    type(struc_Geo_Record) function readGeo_Record(unit_,status)
        integer :: unit_,err,err_alloc,status,status_all,naxis2,frow,felem,uvir
        real :: nullval,anyf
        character nullval_s
        character(len=80) :: comm
        character(len=30) :: errmsg
        type(struc_Geo_Record) :: Geo_Record
        
        ! ------ Init ------
        status_all = 0
        err_alloc = 0
        anyf = 0
        nullval = 0
        nullval_s = '0'
        frow = 1
        felem = 1
        
        ! ------ Move the HDU ------
        call ftmnhd(unit_,2,'Geo_Record',0,status) !ftmhd type 2 = bin table
        status_all = status_all + status
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readGeo_Record: '//errmsg
        endif
        
        ! ------ Array Allocation ------
	! UV/IR
	call ftgkyj(unit_,'NAXIS2',naxis2,comm,status)
	allocate(Geo_Record%number(naxis2),stat=err)
	err_alloc = err_alloc + err
	allocate(Geo_Record%time(naxis2),stat=err)
	err_alloc = err_alloc + err

	if(err_alloc .ne. 0) then
	    print  *,'readGeo_Records: Array allocation error'
	    call exit()
	endif

        ! ------ Read table data ------
	! UV/IR
        call ftgcvs(unit_,1,1,1,naxis2,nullval_s,Geo_Record%time,anyf,status)
	status_all = status_all + status
	call ftgcvj(unit_,2,frow,felem,naxis2,nullval,Geo_Record%number,anyf,status)
	status_all = status_all + status   
		
        status = status_all
        
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readGeo_Record: '//errmsg
        endif
        
        readGeo_Record = Geo_Record
	
    end function readGeo_Record    
    ! ----- readGeo_Record ----------------------------------------------------------------------------------------------------------------------

    
    ! ----- readGeo_Spacecraft ------------------------------------------------------------------------------------------------------------------
    ! Read GEO Spacecraft
    type(struc_Geo_Spacecraft) function readGeo_Spacecraft(unit_,status,uvir)
        integer :: unit_,err,err_alloc,status,status_all,naxis2,frow,felem,uvir
        real :: nullval,anyf
        character nullval_s
        character(len=80) :: comm
        character(len=30) :: errmsg
        type(struc_Geo_Spacecraft) :: Geo_Spacecraft
        
        ! ------ Init ------
        status_all = 0
        err_alloc = 0
        anyf = 0
        nullval = 0
        nullval_s = '0'
        frow = 1
        felem = 1
        
        ! ------ Move the HDU ------
        call ftmnhd(unit_,2,'Geo_Spacecraft',0,status) !ftmhd type 2 = bin table
        status_all = status_all + status
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readGeo_Spacecraft: '//errmsg
        endif
        
	! ------ Array Allocation ------
        ! UV/IR
	call ftgkyj(unit_,'NAXIS2',naxis2,comm,status)
        allocate(Geo_Spacecraft%lat(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Spacecraft%long(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Spacecraft%alt(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Spacecraft%sza(naxis2),stat=err)
        err_alloc = err_alloc + err
        
        ! UV ONLY
        if(uvir .eq. 0) then
		allocate(Geo_Spacecraft%shadowconeratio(naxis2),stat=err)
		err_alloc = err_alloc + err
        endif
        
        if(err_alloc .ne. 0) then
            print  *,'readGeo_Spacecraft: Array allocation error'
            call exit()
        endif        
        
	! ------ Read table data ------
        ! UV/IR
        call ftgcve(unit_,1,frow,felem,naxis2,nullval,Geo_Spacecraft%lat,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,2,frow,felem,naxis2,nullval,Geo_Spacecraft%long,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,3,frow,felem,naxis2,nullval,Geo_Spacecraft%alt,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,4,frow,felem,naxis2,nullval,Geo_Spacecraft%sza,anyf,status)
        status_all = status_all + status
        
        ! UV ONLY
        if(uvir .eq. 0) then
		call ftgcve(unit_,5,frow,felem,naxis2,nullval,Geo_Spacecraft%shadowconeratio,anyf,status)
		status_all = status_all + status
	endif
	
        status = status_all
        
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readGeo_Spacecraft: '//errmsg
        endif
        
        readGeo_Spacecraft = Geo_Spacecraft
        
    end function readGeo_Spacecraft
    ! ----- END readGeo_Spacecraft --------------------------------------------------------------------------------------------------------------
    
    
    ! ----- readGeo_Band1 ----------------------------------------------------------------------------------------------------------------------
    ! Read GEO Band 1
    type(struc_Geo_BandCCDline) function readGeo_Band1(unit_,status,codeop)
        integer :: unit_,err,err_alloc,status,status_all,naxis2,frow,felem,codeop
        real :: nullval,anyf
        character nullval_s
        character(len=80) :: comm
        character(len=30) :: errmsg
        type(struc_Geo_BandCCDline) :: Geo_Band1
        
        ! ------ Init ------
        status_all = 0
        err_alloc = 0
        anyf = 0
        nullval = 0
        nullval_s = '0'
        frow = 1
        felem = 1
        
        ! ------ Array Allocation ------
        call ftgkyj(unit_,'NAXIS2',naxis2,comm,status)
        allocate(Geo_Band1%lat(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band1%long(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band1%alt(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band1%ra(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band1%dec(naxis2),stat=err)
        err_alloc = err_alloc + err    
        
        if(err_alloc .ne. 0) then
            print  *,'readGeo_Band1: Array allocation error'
            call exit()
        endif

        ! ------ Read data ------
        if(codeop .ne. 100) then
            !! Move the HDU to the Geo_Band1 binary table
            call ftmnhd(unit_,2,'Geo_Band1',0,status) !ftmhd type 2 = bin table
            status_all = status_all + status
            if(status .ne. 0) then
                call ftgerr(status,errmsg)
                print *,'readGeo_Band2: '//errmsg
            endif

            call ftgcve(unit_,1,frow,felem,naxis2,nullval,Geo_Band1%lat,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,2,frow,felem,naxis2,nullval,Geo_Band1%long,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,3,frow,felem,naxis2,nullval,Geo_Band1%alt,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,4,frow,felem,naxis2,nullval,Geo_Band1%ra,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,5,frow,felem,naxis2,nullval,Geo_Band1%dec,anyf,status)
            status_all = status_all + status
        else
            Geo_Band1%lat = 0
            Geo_Band1%long = 0
            Geo_Band1%alt = 0
            Geo_Band1%ra = 0
            Geo_Band1%dec = 0
        endif
        
        status = status_all
        
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readGeo_Band1: '//errmsg
        endif
        
        readGeo_Band1 = Geo_Band1            
      
    end function readGeo_Band1
    ! ----- END readGeo_Band1 -----------------------------------------------------------------------------------------------------------------        
       
    
    ! ----- readGeo_Band2 ----------------------------------------------------------------------------------------------------------------------
    ! Read GEO Band 2
    type(struc_Geo_BandCCDline) function readGeo_Band2(unit_,status,codeop)
        integer :: unit_,err,err_alloc,status,status_all,naxis2,frow,felem,codeop
        real :: nullval,anyf
        character nullval_s
        character(len=80) :: comm
        character(len=30) :: errmsg
        type(struc_Geo_BandCCDline) :: Geo_Band2
        
        ! ------ Init ------
        status_all = 0
        err_alloc = 0
        anyf = 0
        nullval = 0
        nullval_s = '0'
        frow = 1
        felem = 1
        
        ! ------ Array Allocation ------
        call ftgkyj(unit_,'NAXIS2',naxis2,comm,status)
        allocate(Geo_Band2%lat(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band2%long(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band2%alt(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band2%ra(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band2%dec(naxis2),stat=err)
        err_alloc = err_alloc + err    
        
        if(err_alloc .ne. 0) then
            print  *,'readGeo_Band2: Array allocation error'
            call exit()
        endif

        ! ------ Read data ------
        if(codeop .ne. 100) then
            !! Move the HDU to the Geo_Band2 binary table
            call ftmnhd(unit_,2,'Geo_Band2',0,status) !ftmhd type 2 = bin table
            status_all = status_all + status
            if(status .ne. 0) then
                call ftgerr(status,errmsg)
                print *,'readGeo_Band2: '//errmsg
            endif

            call ftgcve(unit_,1,frow,felem,naxis2,nullval,Geo_Band2%lat,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,2,frow,felem,naxis2,nullval,Geo_Band2%long,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,3,frow,felem,naxis2,nullval,Geo_Band2%alt,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,4,frow,felem,naxis2,nullval,Geo_Band2%ra,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,5,frow,felem,naxis2,nullval,Geo_Band2%dec,anyf,status)
            status_all = status_all + status
        else
            Geo_Band2%lat = 0
            Geo_Band2%long = 0
            Geo_Band2%alt = 0
            Geo_Band2%ra = 0
            Geo_Band2%dec = 0
        endif
        
        status = status_all

        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readGeo_Band2: '//errmsg
        endif
        
        readGeo_Band2 = Geo_Band2            
            
    end function readGeo_Band2
    ! ----- END readGeo_Band2 -----------------------------------------------------------------------------------------------------------------            
    
    
    ! ----- readGeo_Band3 -----------------------------------------------------------------------------------------------------------------------
    ! Read GEO Band 3
    type(struc_Geo_Band3LOSE) function readGeo_Band3(unit_,status,codeop)
        integer :: unit_,err,err_alloc,status,status_all,naxis2,frow,felem,codeop
        real :: nullval,anyf
        character nullval_s
        character(len=80) :: comm
        character(len=30) :: errmsg
        type(struc_Geo_Band3LOSE) :: Geo_Band3
        
        ! ------ Init ------
        status_all = 0
        err_alloc = 0
        anyf = 0
        nullval = 0
        nullval_s = '0'
        frow = 1
        felem = 1
        
        ! ------ Move the HDU ------
        call ftmnhd(unit_,2,'Geo_Band3',0,status) !ftmhd type 2 = bin table
        status_all = status_all + status
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readGeo_Band3: '//errmsg
        endif
        
        ! ------ Array Allocation ------
        call ftgkyj(unit_,'NAXIS2',naxis2,comm,status)
        allocate(Geo_Band3%sza(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band3%disttoplanetnp(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band3%pixelsize(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band3%anglelossun(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band3%phaselossun(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band3%solarincidence(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band3%solarlocaltime(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band3%anglenormalobserver(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band3%sunazimuth(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band3%distlosplanetcenter(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band3%distscplanetcenter(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band3%lat(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band3%long(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band3%alt(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band3%ra(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band3%dec(naxis2),stat=err)
        err_alloc = err_alloc + err

        if(err_alloc .ne. 0) then
            print  *,'readGeo_Band3: Array allocation error'
            call exit()
        endif

        ! ------ Read table data ------
        if(codeop .ne. 100) then
            call ftgcve(unit_,1,frow,felem,naxis2,nullval,Geo_Band3%lat,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,2,frow,felem,naxis2,nullval,Geo_Band3%long,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,3,frow,felem,naxis2,nullval,Geo_Band3%alt,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,4,frow,felem,naxis2,nullval,Geo_Band3%ra,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,5,frow,felem,naxis2,nullval,Geo_Band3%dec,anyf,status)
            status_all = status_all + status
        else
            Geo_Band3%lat = 0
            Geo_Band3%long = 0
            Geo_Band3%alt = 0
            Geo_Band3%ra = 0
            Geo_Band3%dec = 0
        endif

        call ftgcve(unit_,6,frow,felem,naxis2,nullval,Geo_Band3%sza,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,7,frow,felem,naxis2,nullval,Geo_Band3%disttoplanetnp,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,8,frow,felem,naxis2,nullval,Geo_Band3%pixelsize,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,9,frow,felem,naxis2,nullval,Geo_Band3%anglelossun,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,10,frow,felem,naxis2,nullval,Geo_Band3%phaselossun,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,11,frow,felem,naxis2,nullval,Geo_Band3%solarincidence,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,12,frow,felem,naxis2,nullval,Geo_Band3%solarlocaltime,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,13,frow,felem,naxis2,nullval,Geo_Band3%anglenormalobserver,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,14,frow,felem,naxis2,nullval,Geo_Band3%sunazimuth,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,15,frow,felem,naxis2,nullval,Geo_Band3%distlosplanetcenter,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,16,frow,felem,naxis2,nullval,Geo_Band3%distscplanetcenter,anyf,status)
        status_all = status_all + status
            
        status = status_all
        
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readGeo_Band3: '//errmsg
        endif

        readGeo_Band3 = Geo_Band3            
            
    end function readGeo_Band3
    ! ----- END readGeo_Band3 ------------------------------------------------------------------------------------------------------------------    
    
    
    ! ----- readGeo_Band4 ----------------------------------------------------------------------------------------------------------------------
    ! Read GEO Band 4
    type(struc_Geo_BandCCDline) function readGeo_Band4(unit_,status,codeop)
        integer :: unit_,err,err_alloc,status,status_all,naxis2,frow,felem,codeop
        real :: nullval,anyf
        character nullval_s
        character(len=80) :: comm
        character(len=30) :: errmsg
        type(struc_Geo_BandCCDline) :: Geo_Band4
        
        ! ------ Init ------
        status_all = 0
        err_alloc = 0
        anyf = 0
        nullval = 0
        nullval_s = '0'
        frow = 1
        felem = 1
        
        ! ------ Array Allocation ------
        call ftgkyj(unit_,'NAXIS2',naxis2,comm,status)
        allocate(Geo_Band4%lat(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band4%long(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band4%alt(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band4%ra(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band4%dec(naxis2),stat=err)
        err_alloc = err_alloc + err    
        
        if(err_alloc .ne. 0) then
            print  *,'readGeo_Band4: Array allocation error'
            call exit()
        endif

        ! ------ Read data ------
        if(codeop .ne. 100) then
            !! Move the HDU to the Geo_Band4 binary table
            call ftmnhd(unit_,2,'Geo_Band4',0,status) !ftmhd type 2 = bin table
            status_all = status_all + status
            if(status .ne. 0) then
                call ftgerr(status,errmsg)
                print *,'readGeo_Band4: '//errmsg
            endif

            call ftgcve(unit_,1,frow,felem,naxis2,nullval,Geo_Band4%lat,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,2,frow,felem,naxis2,nullval,Geo_Band4%long,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,3,frow,felem,naxis2,nullval,Geo_Band4%alt,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,4,frow,felem,naxis2,nullval,Geo_Band4%ra,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,5,frow,felem,naxis2,nullval,Geo_Band4%dec,anyf,status)
            status_all = status_all + status
        else
            Geo_Band4%lat = 0
            Geo_Band4%long = 0
            Geo_Band4%alt = 0
            Geo_Band4%ra = 0
            Geo_Band4%dec = 0
        endif
        
        status = status_all
        
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readGeo_Band4: '//errmsg
        endif
        
        readGeo_Band4 = Geo_Band4            
          
    end function readGeo_Band4
    ! ----- END readGeo_Band4 -----------------------------------------------------------------------------------------------------------------                
    
    
    ! ----- readGeo_Band5 ----------------------------------------------------------------------------------------------------------------------
    ! Read GEO Band 5
    type(struc_Geo_BandCCDline) function readGeo_Band5(unit_,status,codeop)
        integer :: unit_,err,err_alloc,status,status_all,naxis2,frow,felem,codeop
        real :: nullval,anyf
        character nullval_s
        character(len=80) :: comm
        character(len=30) :: errmsg
        type(struc_Geo_BandCCDline) :: Geo_Band5
        
        ! ------ Init ------
        status_all = 0
        err_alloc = 0
        anyf = 0
        nullval = 0
        nullval_s = '0'
        frow = 1
        felem = 1
        
        ! ------ Array Allocation ------
        call ftgkyj(unit_,'NAXIS2',naxis2,comm,status)
        allocate(Geo_Band5%lat(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band5%long(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band5%alt(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band5%ra(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Band5%dec(naxis2),stat=err)
        err_alloc = err_alloc + err    
        
        if(err_alloc .ne. 0) then
            print  *,'readGeo_Band5: Array allocation error'
            call exit()
        endif

        ! ------ Read data ------
        if(codeop .ne. 100) then
            !! Move the HDU to the Geo_Band5 binary table
            call ftmnhd(unit_,2,'Geo_Band5',0,status) !ftmhd type 2 = bin table
            status_all = status_all + status
            if(status .ne. 0) then
                call ftgerr(status,errmsg)
                print *,'readGeo_Band5: '//errmsg
            endif

            call ftgcve(unit_,1,frow,felem,naxis2,nullval,Geo_Band5%lat,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,2,frow,felem,naxis2,nullval,Geo_Band5%long,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,3,frow,felem,naxis2,nullval,Geo_Band5%alt,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,4,frow,felem,naxis2,nullval,Geo_Band5%ra,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,5,frow,felem,naxis2,nullval,Geo_Band5%dec,anyf,status)
            status_all = status_all + status
        else
            Geo_Band5%lat = 0
            Geo_Band5%long = 0
            Geo_Band5%alt = 0
            Geo_Band5%ra = 0
            Geo_Band5%dec = 0
        endif
        
        status = status_all
        
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readGeo_Band5: '//errmsg
        endif
        
        readGeo_Band5 = Geo_Band5            
          
    end function readGeo_Band5
    ! ----- END readGeo_Band5 -----------------------------------------------------------------------------------------------------------------                    
    
    
    ! ----- readGeo_CCDLine ----------------------------------------------------------------------------------------------------------------------
    ! Read GEO Geo_CCDLine
    type(struc_Geo_BandCCDline) function readGeo_CCDLine(unit_,status,codeop)
        integer :: unit_,err,err_alloc,status,status_all,naxis2,frow,felem,codeop
        real :: nullval,anyf
        character nullval_s
        character(len=80) :: comm
        character(len=30) :: errmsg
        type(struc_Geo_BandCCDline) :: Geo_CCDLine
        
        ! ------ Init ------
        status_all = 0
        err_alloc = 0
        anyf = 0
        nullval = 0
        nullval_s = '0'
        frow = 1
        felem = 1
        
        ! ------ Array Allocation ------
        call ftgkyj(unit_,'NAXIS2',naxis2,comm,status)
        allocate(Geo_CCDLine%lat(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_CCDLine%long(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_CCDLine%alt(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_CCDLine%ra(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_CCDLine%dec(naxis2),stat=err)
        err_alloc = err_alloc + err    
        
        if(err_alloc .ne. 0) then
            print  *,'readGeo_CCDLine: Array allocation error'
            call exit()
        endif

        ! ------ Read data ------
        if(codeop .eq. 100) then
            !! Move the HDU to the Geo_CCDLine binary table
            call ftmnhd(unit_,2,'Geo_CCDLine',0,status) !ftmhd type 2 = bin table
            status_all = status_all + status
            if(status .ne. 0) then
                call ftgerr(status,errmsg)
                print *,'readGeo_CCDLine: '//errmsg
            endif

            call ftgcve(unit_,1,frow,felem,naxis2,nullval,Geo_CCDLine%lat,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,2,frow,felem,naxis2,nullval,Geo_CCDLine%long,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,3,frow,felem,naxis2,nullval,Geo_CCDLine%alt,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,4,frow,felem,naxis2,nullval,Geo_CCDLine%ra,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,5,frow,felem,naxis2,nullval,Geo_CCDLine%dec,anyf,status)
            status_all = status_all + status
        else
            Geo_CCDLine%lat = 0
            Geo_CCDLine%long = 0
            Geo_CCDLine%alt = 0
            Geo_CCDLine%ra = 0
            Geo_CCDLine%dec = 0
        endif
        
        status = status_all
        
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readGeo_CCDLine: '//errmsg
        endif

        readGeo_CCDLine = Geo_CCDLine            
           
    end function readGeo_CCDLine
    ! ----- END readGeo_CCDLine -----------------------------------------------------------------------------------------------------------------                        
    
    
    ! ----- readGeo_LOSE ------------------------------------------------------------------------------------------------------------------------
    ! Read GEO Geo_LOSE
    type(struc_Geo_Band3LOSE) function readGeo_LOSE(unit_,status)
        integer :: unit_,err,err_alloc,status,status_all,naxis2,frow,felem
        real :: nullval,anyf
        character nullval_s
        character(len=80) :: comm
        character(len=30) :: errmsg
        type(struc_Geo_Band3LOSE) :: Geo_LOSE
        
        ! ------ Init ------
        status_all = 0
        err_alloc = 0
        anyf = 0
        nullval = 0
        nullval_s = '0'
        frow = 1
        felem = 1
        
        ! ------ Array Allocation ------
        call ftgkyj(unit_,'NAXIS2',naxis2,comm,status)
        allocate(Geo_LOSE%sza(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_LOSE%disttoplanetnp(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_LOSE%pixelsize(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_LOSE%anglelossun(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_LOSE%phaselossun(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_LOSE%solarincidence(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_LOSE%solarlocaltime(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_LOSE%anglenormalobserver(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_LOSE%sunazimuth(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_LOSE%distlosplanetcenter(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_LOSE%distscplanetcenter(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_LOSE%lat(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_LOSE%long(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_LOSE%alt(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_LOSE%ra(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_LOSE%dec(naxis2),stat=err)
        err_alloc = err_alloc + err

        if(err_alloc .ne. 0) then
            print  *,'readGeo: Array allocation error'
            call exit()
        endif

        ! ------ Read data ------
        !! Move the HDU to the Geo_LOSE binary table, if existing
        call ftmnhd(unit_,2,'Geo_LOSE',0,status) !ftmhd type 2 = bin table
        
        if(status .eq. 0) then
            call ftgcve(unit_,1,frow,felem,naxis2,nullval,Geo_LOSE%lat,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,2,frow,felem,naxis2,nullval,Geo_LOSE%long,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,3,frow,felem,naxis2,nullval,Geo_LOSE%alt,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,4,frow,felem,naxis2,nullval,Geo_LOSE%ra,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,5,frow,felem,naxis2,nullval,Geo_LOSE%dec,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,6,frow,felem,naxis2,nullval,Geo_LOSE%sza,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,7,frow,felem,naxis2,nullval,Geo_LOSE%disttoplanetnp,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,8,frow,felem,naxis2,nullval,Geo_LOSE%pixelsize,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,9,frow,felem,naxis2,nullval,Geo_LOSE%anglelossun,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,10,frow,felem,naxis2,nullval,Geo_LOSE%phaselossun,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,11,frow,felem,naxis2,nullval,Geo_LOSE%solarincidence,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,12,frow,felem,naxis2,nullval,Geo_LOSE%solarlocaltime,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,13,frow,felem,naxis2,nullval,Geo_LOSE%anglenormalobserver,anyf,status)
            status_all = status_all + status
            call ftgcve(unit_,14,frow,felem,naxis2,nullval,Geo_LOSE%sunazimuth,anyf,status)
            status_all = status_all + status
            !call ftgcve(unit_,15,frow,felem,naxis2,nullval,Geo_LOSE%distlosplanetcenter,anyf,status)
            !status_all = status_all + status
            !call ftgcve(unit_,16,frow,felem,naxis2,nullval,Geo_LOSE%distscplanetcenter,anyf,status)
            !status_all = status_all + status
            Geo_LOSE%distlosplanetcenter = 0
            Geo_LOSE%distscplanetcenter = 0
        else
            Geo_LOSE%lat = 0
            Geo_LOSE%long = 0
            Geo_LOSE%alt = 0
            Geo_LOSE%ra = 0
            Geo_LOSE%dec = 0
            Geo_LOSE%sza = 0
            Geo_LOSE%disttoplanetnp = 0
            Geo_LOSE%pixelsize = 0
            Geo_LOSE%anglelossun = 0
            Geo_LOSE%phaselossun = 0
            Geo_LOSE%solarincidence = 0
            Geo_LOSE%solarlocaltime = 0
            Geo_LOSE%anglenormalobserver = 0
            Geo_LOSE%sunazimuth = 0
            Geo_LOSE%distlosplanetcenter = 0
            Geo_LOSE%distscplanetcenter = 0
        endif
        
        status = status_all
        
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readGeo_LOSE: '//errmsg
        endif
        
        readGeo_LOSE = Geo_LOSE          
           
    end function readGeo_LOSE
    ! ----- END readGeo_LOSE --------------------------------------------------------------------------------------------------------------------
    
    
    ! ----- readGeo_IRFOV -----------------------------------------------------------------------------------------------------------------------
    ! Read GEO IRFOV
    type(struc_Geo_Band3LOSE) function readGeo_IRFOV(unit_,status)
        integer :: unit_,err,err_alloc,status,status_all,naxis2,frow,felem
        real :: nullval,anyf
        character nullval_s
        character(len=80) :: comm
        character(len=30) :: errmsg
        type(struc_Geo_Band3LOSE) :: Geo_IRFOV
        
        ! ------ Init ------
        status_all = 0
        err_alloc = 0
        anyf = 0
        nullval = 0
        nullval_s = '0'
        frow = 1
        felem = 1
        
        ! ------ Move the HDU ------
        call ftmnhd(unit_,2,'Geo_IRFOV',0,status) !ftmhd type 2 = bin table
        status_all = status_all + status
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readGeo_IRFOV: '//errmsg
        endif
        
        ! ------ Array Allocation ------
        call ftgkyj(unit_,'NAXIS2',naxis2,comm,status)
        allocate(Geo_IRFOV%sza(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_IRFOV%disttoplanetnp(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_IRFOV%pixelsize(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_IRFOV%anglelossun(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_IRFOV%phaselossun(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_IRFOV%solarincidence(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_IRFOV%solarlocaltime(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_IRFOV%anglenormalobserver(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_IRFOV%sunazimuth(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_IRFOV%distlosplanetcenter(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_IRFOV%distscplanetcenter(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_IRFOV%lat(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_IRFOV%long(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_IRFOV%alt(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_IRFOV%ra(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_IRFOV%dec(naxis2),stat=err)
        err_alloc = err_alloc + err

        if(err_alloc .ne. 0) then
            print  *,'readGeo_IRFOV: Array allocation error'
            call exit()
        endif

        ! ------ Read table data ------
	call ftgcve(unit_,1,frow,felem,naxis2,nullval,Geo_IRFOV%lat,anyf,status)
	status_all = status_all + status
	call ftgcve(unit_,2,frow,felem,naxis2,nullval,Geo_IRFOV%long,anyf,status)
	status_all = status_all + status
	call ftgcve(unit_,3,frow,felem,naxis2,nullval,Geo_IRFOV%alt,anyf,status)
	status_all = status_all + status
	call ftgcve(unit_,4,frow,felem,naxis2,nullval,Geo_IRFOV%ra,anyf,status)
	status_all = status_all + status
	call ftgcve(unit_,5,frow,felem,naxis2,nullval,Geo_IRFOV%dec,anyf,status)
	status_all = status_all + status
        call ftgcve(unit_,6,frow,felem,naxis2,nullval,Geo_IRFOV%sza,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,7,frow,felem,naxis2,nullval,Geo_IRFOV%disttoplanetnp,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,8,frow,felem,naxis2,nullval,Geo_IRFOV%pixelsize,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,9,frow,felem,naxis2,nullval,Geo_IRFOV%anglelossun,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,10,frow,felem,naxis2,nullval,Geo_IRFOV%phaselossun,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,11,frow,felem,naxis2,nullval,Geo_IRFOV%solarincidence,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,12,frow,felem,naxis2,nullval,Geo_IRFOV%solarlocaltime,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,13,frow,felem,naxis2,nullval,Geo_IRFOV%anglenormalobserver,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,14,frow,felem,naxis2,nullval,Geo_IRFOV%sunazimuth,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,15,frow,felem,naxis2,nullval,Geo_IRFOV%distlosplanetcenter,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,16,frow,felem,naxis2,nullval,Geo_IRFOV%distscplanetcenter,anyf,status)
        status_all = status_all + status
            
        status = status_all
        
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readGeo_IRFOV: '//errmsg
        endif
        
        readGeo_IRFOV = Geo_IRFOV            
            
    end function readGeo_IRFOV
    ! ----- END readGeo_IRFOV -------------------------------------------------------------------------------------------------------------------
    
    
    
    ! ----- readGeo_Coordinates -----------------------------------------------------------------------------------------------------------------
    ! Read GEO Geo_Coordinates
    type(struc_Geo_Coordinates) function readGeo_Coordinates(unit_,status)
        integer :: unit_,err,err_alloc,status,status_all,naxis2,frow,felem
        real :: nullval,anyf
        character nullval_s
        character(len=80) :: comm
        character(len=30) :: errmsg
        type(struc_Geo_Coordinates) :: Geo_Coordinates
        
        ! ------ Init ------
        status_all = 0
        err_alloc = 0
        anyf = 0
        nullval = 0
        nullval_s = '0'
        frow = 1
        felem = 1
        
        ! ------ Move the HDU ------
        call ftmnhd(unit_,2,'Geo_Coordinates',0,status) !ftmhd type 2 = bin table
        status_all = status_all + status
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readGeo_Coordinates: '//errmsg
        endif
        
        ! ------ Array Allocation ------
        call ftgkyj(unit_,'NAXIS2',naxis2,comm,status)
        allocate(Geo_Coordinates%xsc_x(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Coordinates%xsc_y(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Coordinates%xsc_z(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Coordinates%ysc_x(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Coordinates%ysc_y(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Coordinates%ysc_z(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Coordinates%zsc_x(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Coordinates%zsc_y(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Coordinates%zsc_z(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Coordinates%angle_xsc_localvert(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Coordinates%angle_xsc_proj(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Coordinates%p1_u(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Coordinates%p1_v(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Coordinates%p2_u(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Coordinates%p2_v(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Coordinates%p3_u(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Coordinates%p3_v(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Coordinates%p4_u(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Coordinates%p4_v(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Coordinates%p5_u(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_Coordinates%p5_v(naxis2),stat=err)
        err_alloc = err_alloc + err

        if(err_alloc .ne. 0) then
            print  *,'readGeo_Coordinates: Array allocation error'
            call exit()
        endif

        ! ------ Read data ------
        call ftgcve(unit_,1,frow,felem,naxis2,nullval,Geo_Coordinates%xsc_x,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,2,frow,felem,naxis2,nullval,Geo_Coordinates%xsc_y,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,3,frow,felem,naxis2,nullval,Geo_Coordinates%xsc_z,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,4,frow,felem,naxis2,nullval,Geo_Coordinates%ysc_x,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,5,frow,felem,naxis2,nullval,Geo_Coordinates%ysc_y,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,6,frow,felem,naxis2,nullval,Geo_Coordinates%ysc_z,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,7,frow,felem,naxis2,nullval,Geo_Coordinates%zsc_x,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,8,frow,felem,naxis2,nullval,Geo_Coordinates%zsc_y,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,9,frow,felem,naxis2,nullval,Geo_Coordinates%zsc_z,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,10,frow,felem,naxis2,nullval,Geo_Coordinates%angle_xsc_localvert,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,11,frow,felem,naxis2,nullval,Geo_Coordinates%angle_xsc_proj,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,12,frow,felem,naxis2,nullval,Geo_Coordinates%p1_u,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,13,frow,felem,naxis2,nullval,Geo_Coordinates%p1_v,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,14,frow,felem,naxis2,nullval,Geo_Coordinates%p2_u,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,15,frow,felem,naxis2,nullval,Geo_Coordinates%p2_v,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,16,frow,felem,naxis2,nullval,Geo_Coordinates%p3_u,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,17,frow,felem,naxis2,nullval,Geo_Coordinates%p3_v,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,18,frow,felem,naxis2,nullval,Geo_Coordinates%p4_u,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,19,frow,felem,naxis2,nullval,Geo_Coordinates%p4_v,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,20,frow,felem,naxis2,nullval,Geo_Coordinates%p5_u,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,21,frow,felem,naxis2,nullval,Geo_Coordinates%p5_v,anyf,status)
        status_all = status_all + status
        
        status = status_all
        
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readGeo_Coordinates: '//errmsg
        endif
        
        readGeo_Coordinates = Geo_Coordinates          
            
    end function readGeo_Coordinates
    ! ----- END readGeo_Coordinates -------------------------------------------------------------------------------------------------------------
    
    
    ! ----- readGeo_TransMatrix -----------------------------------------------------------------------------------------------------------------
    ! Read GEO Geo_TransMatrix
    type(struc_Geo_TransMatrix) function readGeo_TransMatrix(unit_,status)
        integer :: unit_,err,err_alloc,status,status_all,naxis2,frow,felem
        real :: nullval,anyf
        character nullval_s
        character(len=80) :: comm
        character(len=30) :: errmsg
        type(struc_Geo_TransMatrix) :: Geo_TransMatrix
        
        ! ------ Init ------
        status_all = 0
        err_alloc = 0
        anyf = 0
        nullval = 0
        nullval_s = '0'
        frow = 1
        felem = 1
        
        ! ------ Move the HDU ------
        call ftmnhd(unit_,2,'Geo_TransMatrix',0,status) !ftmhd type 2 = bin table
        status_all = status_all + status
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readGeo_TransMatrix: '//errmsg
        endif
        
        ! ------ Array Allocation ------
        call ftgkyj(unit_,'NAXIS2',naxis2,comm,status)    
        allocate(Geo_TransMatrix%x_ra(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_TransMatrix%x_dec(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_TransMatrix%y_ra(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_TransMatrix%y_dec(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_TransMatrix%z_ra(naxis2),stat=err)
        err_alloc = err_alloc + err
        allocate(Geo_TransMatrix%z_dec(naxis2),stat=err)
        err_alloc = err_alloc + err

        if(err_alloc .ne. 0) then
            print  *,'readGeo: Array allocation error'
            call exit()
        endif

        call ftgcve(unit_,1,frow,felem,naxis2,nullval,Geo_TransMatrix%x_ra,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,2,frow,felem,naxis2,nullval,Geo_TransMatrix%x_dec,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,3,frow,felem,naxis2,nullval,Geo_TransMatrix%y_ra,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,4,frow,felem,naxis2,nullval,Geo_TransMatrix%y_dec,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,5,frow,felem,naxis2,nullval,Geo_TransMatrix%z_ra,anyf,status)
        status_all = status_all + status
        call ftgcve(unit_,5,frow,felem,naxis2,nullval,Geo_TransMatrix%z_dec,anyf,status)
        status_all = status_all + status

        status = status_all
        
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readGeo_TransMatrix: '//errmsg
        endif
        
        readGeo_TransMatrix = Geo_TransMatrix                  
        
    end function readGeo_TransMatrix
    ! ----- END readGeo_TransMatrix -------------------------------------------------------------------------------------------------------------
        
        
    ! ----- readGeo -----------------------------------------------------------------------------------------------------------------------------
    ! Read GEO
    type(struc_Geo) function readGeo(unit_,status,codeop,mode,uvir)    
        integer :: unit_,status,status_all,codeop,mode,uvir
        character(len=30) :: errmsg
        type(struc_Geo) :: Geo 
	type(struc_Geo_Record) :: Geo_Record
	type(struc_Geo_Spacecraft) :: Geo_Spacecraft        
	type(struc_Geo_Coordinates) :: Geo_Coordinates
	type(struc_Geo_TransMatrix) :: Geo_TransMatrix	
	type(struc_Geo_BandCCDline) :: Geo_Band1
	type(struc_Geo_BandCCDline) :: Geo_Band2
	type(struc_Geo_Band3LOSE) :: Geo_Band3
	type(struc_Geo_BandCCDline) :: Geo_Band4
	type(struc_Geo_BandCCDline) :: Geo_Band5
	type(struc_Geo_BandCCDline) :: Geo_CCDLine
	type(struc_Geo_Band3LOSE) :: Geo_LOSE
	type(struc_Geo_Band3LOSE) :: Geo_IRFOV
	
        ! ------ Init ------
        status_all = 0
        
        ! ------ Build GEO structure ------
        ! UV/IR
	Geo_Record = readGeo_Record(unit_,status)
	if(status .eq. 0) then
		if(mode .ne. 0) then
			print *,'Geo_Record successfully read'
		endif
		Geo%Record = Geo_Record
		status_all = status_all + status
	else
		call ftgerr(status,errmsg)
		print *,'readGeo_Record: '//errmsg
	endif
	Geo_Spacecraft = readGeo_Spacecraft(unit_,status,uvir)
	if(status .eq. 0) then
		if(mode .ne. 0) then
			print *,'Geo_Spacecraft successfully read'
		endif
		Geo%Spacecraft = Geo_Spacecraft
		status_all = status_all + status
	else
		call ftgerr(status,errmsg)
		print *,'readGeo_Spacecraft: '//errmsg
	endif
	Geo_Coordinates = readGeo_Coordinates(unit_,status)
	if(status .eq. 0) then
		if(mode .ne. 0) then
			print *,'Geo_Coordinates successfully read'
		endif
		Geo%Coordinates = Geo_Coordinates
		status_all = status_all + status
	else
		call ftgerr(status,errmsg)
		print *,'readGeo_Coordinates: '//errmsg
	endif
	Geo_TransMatrix = readGeo_TransMatrix(unit_,status)
	if(status .eq. 0) then
		if(mode .ne. 0) then
			print *,'Geo_TransMatrix successfully read'
		endif
		Geo%TransMatrix = Geo_TransMatrix
		status_all = status_all + status
	endif    

        ! UV ONLY
        if(uvir .eq. 0) then
		Geo_Band1 = readGeo_Band1(unit_,status,codeop)
		if(status .eq. 0) then
			if(mode .ne. 0) then
				print *,'Geo_Band1 successfully read'
			endif
			Geo%Band1 = Geo_Band1
			status_all = status_all + status
		endif
        
		Geo_Band2 = readGeo_Band2(unit_,status,codeop)
		if(status .eq. 0) then
			if(mode .ne. 0) then
				print *,'Geo_Band2 successfully read'
			endif
			Geo%Band2 = Geo_Band2
			status_all = status_all + status
		endif
		
		Geo_Band3 = readGeo_Band3(unit_,status,codeop)
		if(status .eq. 0) then
			if(mode .ne. 0) then
				print *,'Geo_Band3 successfully read'
			endif
			Geo%Band3 = Geo_Band3
			status_all = status_all + status
		endif		
		
		Geo_Band4 = readGeo_Band4(unit_,status,codeop)
		if(status .eq. 0) then
			if(mode .ne. 0) then
				print *,'Geo_Band4 successfully read'
			endif
			Geo%Band4 = Geo_Band4
			status_all = status_all + status
		endif
		
		Geo_Band5 = readGeo_Band5(unit_,status,codeop)
		if(status .eq. 0) then
			if(mode .ne. 0) then
				print *,'Geo_Band5 successfully read'
			endif
			Geo%Band5 = Geo_Band5
			status_all = status_all + status
		endif			
		
		Geo_CCDLine = readGeo_CCDLine(unit_,status,codeop)
		if(status .eq. 0) then
			if(mode .ne. 0) then
				print *,'Geo_CCDLine successfully read'
			endif
			Geo%CCDLine = Geo_CCDLine
			status_all = status_all + status
		endif					

		Geo_LOSE = readGeo_LOSE(unit_,status)
		if(status .eq. 0) then
			if(mode .ne. 0) then
				print *,'Geo_LOSE successfully read'
			endif
			Geo%LOSE = Geo_LOSE
			status_all = status_all + status
		endif							
        endif    
        
        ! IR ONLY
        if(uvir .eq. 1) then
		Geo_IRFOV = readGeo_IRFOV(unit_,status)
		if(status .eq. 0) then
			if(mode .ne. 0) then
				print *,'Geo_IRFOV successfully read'
			endif
			Geo%IRFOV = Geo_IRFOV
			status_all = status_all + status
		endif							        
        endif 
        
        status = status_all

        readGeo = Geo                  
        
    end function readGeo
    ! ----- END readGeo -------------------------------------------------------------------------------------------------------------------------

    
    ! ----- readCleanData -----------------------------------------------------------------------------------------------------------------------
    ! Read Clean Data
    subroutine readCleanData(unit_,status,Naxis1,Naxis2,Naxis3,Data)
        integer :: unit_,status,type,status_all
        integer :: Naxis1,Naxis2,Naxis3
        real :: nullval = 0, anyf=0
        real, dimension(Naxis1,Naxis2,Naxis3) :: Data
        character(len=30) errmsg

        ! Read the primary Data
        call ftmahd(unit_,1,type,status)
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readCleanData: '//errmsg
        endif
        status_all = status

        call ftg3de(unit_,1,nullval,Naxis1,Naxis2,Naxis1,Naxis2,Naxis3,Data,anyf,status)
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readCleanData: '//errmsg
        endif
        status_all = status_all + status

        status = status_all

    end subroutine readCleanData
    ! ----- END readCleanData -------------------------------------------------------------------------------------------------------------------

    
    ! ----- readFlag ----------------------------------------------------------------------------------------------------------------------------
    ! Read Image extension Flag
    subroutine readFlag(unit_,status,Naxis1,Naxis2,Naxis3,Flag)
        integer :: unit_,status,status_all
        integer :: Naxis1,Naxis2,Naxis3
        integer, dimension(Naxis1,Naxis2,Naxis3) :: Flag
        real :: nullval = 0, anyf=0
        character(len=30) errmsg

        ! Move the HDU to the Flag image extension
        call ftmnhd(unit_,0,'Flag',0,status) !ftmhd type 0 = img extension
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readFlag: '//errmsg
        endif
        status_all = status

        ! Read the image extension
        call ftg3dj(unit_,1,nullval,Naxis1,Naxis2,Naxis1,Naxis2,Naxis3,Flag,anyf,status)
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readFlag: '//errmsg
        endif
        status_all = status_all + status

        status = status_all

    end subroutine readFlag
    ! ----- END readFlag ------------------------------------------------------------------------------------------------------------------------
    
    
    ! ----- applyFlag ---------------------------------------------------------------------------------------------------------------------------
    ! Apply Flag
    subroutine applyFlag(status,Naxis1,Naxis2,Naxis3,Flag,CleanData)
        integer :: status,Naxis1,Naxis2,Naxis3
        real, dimension(Naxis1,Naxis2,Naxis3) :: CleanData
        integer, dimension(Naxis1,Naxis2,Naxis3) :: Flag

        ! Set to NaN where the Flag is equal to 1/2/3/4
        where((Flag .eq. 1) .or. (Flag .eq. 2) .or. (Flag .eq. 3) .or. (Flag .eq. 4))
            CleanData = setnan(-1.)
        end where

        status = 0

    end subroutine applyFlag
    ! ----- END applyFlag -----------------------------------------------------------------------------------------------------------------------
    

    ! ----- setnan ------------------------------------------------------------------------------------------------------------------------------
    ! Set to NaN
    real function setnan(arg)
        real arg
        setnan = sqrt(arg)
    end function setnan
    ! ----- END setnan --------------------------------------------------------------------------------------------------------------------------
    
    
    ! ----- readErrData -------------------------------------------------------------------------------------------------------------------------
    ! Read Image extension ErrData
    subroutine readErrData(unit_,status,Naxis1,Naxis2,Naxis3,ErrData)
        integer :: unit_,status,status_all
        integer :: Naxis1,Naxis2,Naxis3
        real, dimension(Naxis1,Naxis2,Naxis3) :: ErrData
        real :: nullval = 0, anyf=0
        character(len=30) errmsg

        ! Move the HDU to the Flag image extension
        call ftmnhd(unit_,0,'ErrData',0,status) !ftmhd type 0 = img extension
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readErrData: '//errmsg
        endif
        status_all = status

        ! Read the image extension
        call ftg3de(unit_,1,nullval,Naxis1,Naxis2,Naxis1,Naxis2,Naxis3,ErrData,anyf,status)
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readErrData: '//errmsg
        endif
        status_all = status_all + status

        status = status_all

    end subroutine readErrData
    ! ----- END readErrData ---------------------------------------------------------------------------------------------------------------------
    
    
    ! ----- file_basename -----------------------------------------------------------------------------------------------------------------------
    ! Get the basename  of a filename
    character(len=32) function file_basename(fname)
        integer pos
        character(len=2048) fname

        pos = index(fname,"/",.true.)
        file_basename = trim(fname(pos+1:len(fname)))

    end function file_basename
    ! ----- END file_basename -------------------------------------------------------------------------------------------------------------------
    
    
    ! ----- readWL ------------------------------------------------------------------------------------------------------------------------------
    ! Read Image extension WAVELENGTH
    subroutine readWL(unit_,status,Naxis1,Naxis2,Naxis3,WL)
        integer :: unit_,status,status_all
        integer :: Naxis1,Naxis2,Naxis3
        real, dimension(:,:,:), allocatable :: WL
        real :: nullval = 0, anyf=0
        character(len=30) errmsg

        ! Move the HDU to the WAVELENGTH image extension
        call ftmnhd(unit_,0,'WAVELENGTH',0,status) !ftmhd type 0 = img extension
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readWL: '//errmsg
        endif
        status_all = status

        ! Read the image extension
        call ftg3de(unit_,1,nullval,Naxis1,Naxis2,Naxis1,Naxis2,Naxis3,WL,anyf,status)
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readWL: '//errmsg
        endif
        status_all = status_all + status

        status = status_all

    end subroutine readWL
    ! ----- END readErrData ---------------------------------------------------------------------------------------------------------------------


    ! ----- readDC ------------------------------------------------------------------------------------------------------------------------------
    ! Read Image extension DC
    subroutine readDC(unit_,status,Naxis1,Naxis2,Naxis3,DC)
        integer :: unit_,status,status_all
        integer :: Naxis1,Naxis2,Naxis3
        real, dimension(:,:,:), allocatable :: DC
        real :: nullval = 0, anyf=0
        character(len=30) errmsg

        ! Move the HDU to the DC image extension
        call ftmnhd(unit_,0,'DC',0,status) !ftmhd type 0 = img extension
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readDC: '//errmsg
        endif
        status_all = status

        ! Read the image extension
        call ftg3de(unit_,1,nullval,Naxis1,Naxis2,Naxis1,Naxis2,Naxis3,DC,anyf,status)
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readDC: '//errmsg
        endif
        status_all = status_all + status

        status = status_all

    end subroutine readDC
    ! ----- END readDC --------------------------------------------------------------------------------------------------------------------------
    
    
    ! ----- readRaw -----------------------------------------------------------------------------------------------------------------------------
    ! Read Image extension RAW
    subroutine readRaw(unit_,status,Naxis1,Naxis2,Naxis3,RAW)
        integer :: unit_,status,status_all
        integer :: Naxis1,Naxis2,Naxis3
        real, dimension(:,:,:), allocatable :: RAW
        real :: nullval = 0, anyf=0
        character(len=30) errmsg

        ! Move the HDU to the RAW image extension
        call ftmnhd(unit_,0,'RAW',0,status) !ftmhd type 0 = img extension
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readRaw: '//errmsg
        endif
        status_all = status

        ! Read the image extension
        call ftg3de(unit_,1,nullval,Naxis1,Naxis2,Naxis1,Naxis2,Naxis3,RAW,anyf,status)
        if(status .ne. 0) then
            call ftgerr(status,errmsg)
            print *,'readRaw: '//errmsg
        endif
        status_all = status_all + status

        status = status_all

    end subroutine readRaw
    ! ----- END readRaw -------------------------------------------------------------------------------------------------------------------------

    
! ----- Functions ----------------------------------------------------------------------------------------------------------------------------------------------

    
end module spica_struct_1a_mod_func

